/* Action/Movement refers to a whole class of movement-related actions.

   Action/Movement by itself is the main one simply used as a move_towards target action.
     (Eventually, it should even attempt to use pathing.)

   Nested within Action/Movement are special movement modes such as the fine movement used by player.

   All Action/Movement procs should have this in common

   var/target - the actual turf that is the destination for this movement.
   var/what - the direction you hope to be facing for this movement.
*/


Action/Movement

	New(var/newActor,var/newTarget,var/newModifier)

		actor = newActor
		target = newTarget
		if (modifier != null)
			modifier = newModifier
		else
			modifier = GetLastStep()

	/* GenerateStepNoiseOldAndNewTurf
		Generates two steps, one at 3 ms at the turf that was just left, and one at 8 ms at the turf that was just entered. */
	proc/GenerateStepNoiseOldAndNewTurf(var/turf/oldTurf,var/turf/newTurf)

		var/const/FIRST_STEP_DELAY = 2 // Animation makes it look like 3
		var/const/SECOND_STEP_DELAY = 7 // Animation makes it look like 8
		var/const/PREFERRED_STEP_VOLUME = 50

		if (!istype(oldTurf) || !istype(newTurf))
			DebugMsg("[type].GenerateStepNoiseOldAndNewTurf() requires turfs.")
		else
			spawn()
				sleep(3)
				oldTurf.MakeTurfNoise(PREFERRED_STEP_VOLUME)
			spawn()
				sleep(8)
				newTurf.MakeTurfNoise(PREFERRED_STEP_VOLUME)




	/* GetLastStep()
	Goes through the actionQueue, backwards first, and returns the last step taken.
	It relies on the last step being stored in the actionExamined.subject.
	If no movement orders are encountered, it assumes the mob's current location is the last step.*/
	proc/GetLastStep()

		var/tmp/turf/lastStep = null

		var/Action/actionExamined

		lastStep = null

		for (var/thisIncrement = actor.actions.len, thisIncrement > 0, thisIncrement--)
			actionExamined = actor.actions[thisIncrement]

			if (istype(actionExamined,/Action/Movement))
				lastStep = actionExamined.target

			if (lastStep != null) return (lastStep)
			continue

		return (actor.loc)

	/* GetLastDir()
	Goes through the actionQueue, bkacwards first, and returns the last direction taken.
	It relies on the last direction being stored on actionExamined.what */
	proc/GetLastDir()

		var/tmp/lastDir = null

		var/tmp/Action/actionExamined

		lastDir = null

		for (var/thisIncrement = actor.actions.len, thisIncrement > 0, thisIncrement--)
			actionExamined = actor.actions[thisIncrement]

			if (istype(actionExamined,/Action/Movement))
				lastDir = actionExamined.modifier

			if (lastDir != null) return (lastDir)

		return (actor.dir)

	/* StepMovement()
		All fine step movements are identical code, the main difference simply being in their initialization and
		the resulting population of target and modifier. */
	proc/StepMovement()

		if (target == null)
			return(ACTION_STATUS_IMPOSSIBLE)

		var/turf/oldTurf = actor.loc

		if (actor.Move(target,modifier))

			GenerateStepNoiseOldAndNewTurf(oldTurf,actor.loc)

			// The reason I'm going with this is in case there's no "idle", but I could see it backfiring.

			if ((type == /Action/Movement/StepRight) || (type == /Action/Movement/StepLeft))

				if ("sidestep" in icon_states(actor.icon))

					actor.icon_state = "sidestep"

			else
				if ("moving" in icon_states(actor.icon))

					actor.icon_state = "moving"

			// Here's where the time related to movement would go.

			actor.ActionMoved()

			if ("idle" in icon_states(actor.icon))

				actor.icon_state = "idle"

			return(ACTION_STATUS_COMPLETE)

		else

			return(ACTION_STATUS_IMPOSSIBLE)

	/* RemoveStepIfInvalid()
		All fine step movements are checked against this.
		If all goes poorly, the movement is removed from the queue. */
	proc/RemoveStepIfInvalid()
		var/turf/thisTurf = target
		if (!thisTurf.Enter(actor))
			actor.GameOutput("You are unable to move there.")
			actor.RemoveActionFromQueue(src)
			del src

	/* RemoveRedundantMovesIfFound()
		This proc goes backwards through the action queue looking specifically for
		conflicting moves such as turning right immediately after turning left and pares the difference.
		It can be used to reverse queued moves. */

		// If I don't keep track of this, I end up going back through the whole queue, which may be unnecessary.
	var/const/MAXIMUM_MOVES_TO_CHECK_IF_REDUNANT = 4

	proc/RemoveRedundantMovesIfFound()
		if (actor.actions.len < 2)
			return()

		var/matchFound = 1

		while (matchFound == 1)
			for (var/increment = actor.actions.len,increment > 1,increment--)

				if (increment < (actor.actions.len - MAXIMUM_MOVES_TO_CHECK_IF_REDUNANT))
					return()

				var/Action/ActionExamined = actor.actions[increment]
				var/Action/priorActionExamined = actor.actions[increment-1]

				if (istype(ActionExamined,/Action/Movement/TurnLeft) && istype(priorActionExamined,/Action/Movement/TurnRight))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
				else if (istype(ActionExamined,/Action/Movement/TurnRight) && istype(priorActionExamined,/Action/Movement/TurnLeft))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
				if (istype(ActionExamined,/Action/Movement/StepForward) && istype(priorActionExamined,/Action/Movement/StepBackward))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
				else if (istype(ActionExamined,/Action/Movement/StepBackward) && istype(priorActionExamined,/Action/Movement/StepForward))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
				else if (istype(ActionExamined,/Action/Movement/StepRight) && istype(priorActionExamined,/Action/Movement/StepLeft))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
				else if (istype(ActionExamined,/Action/Movement/StepLeft) && istype(priorActionExamined,/Action/Movement/StepRight))
					actor.RemoveActionFromQueue(ActionExamined)
					actor.RemoveActionFromQueue(priorActionExamined)
					matchFound = 1
					break
			matchFound = 0


/* The following are fine movement for player/mobs. */

Action/Movement/StepForward

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		modifier = GetLastDir()
		target = get_step(GetLastStep(),modifier)

		RemoveStepIfInvalid()
		RemoveRedundantMovesIfFound()

	Attempt()

		return(StepMovement())

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		if (target != null)
			var/turf/turfPriorToMove = get_step(target,turn(modifier,180))
			if (turfPriorToMove != null)
				var/imageNewImage = image('imgControlScheme.dmi',turfPriorToMove,"forward",3,modifier)
				ShiftImageInItsDirection(imageNewImage)
				myImages.Add(imageNewImage)
				return(myImages)
		return(null)

Action/Movement/StepBackward

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		modifier = GetLastDir() // We're not turning 180 here, we're facing the same direction, just turning.
		target = get_step(GetLastStep(),turn(modifier,180))

		RemoveStepIfInvalid()
		RemoveRedundantMovesIfFound()

	Attempt()

		return(StepMovement())

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		if (target != null)
			var/turf/turfPriorToMove = get_step(target,modifier)
			if (turfPriorToMove != null)
				var/imageNewImage = image('imgControlScheme.dmi',turfPriorToMove,"backward",3,modifier)
				ShiftImageInItsOppositeDirection(imageNewImage)
				myImages.Add(imageNewImage)
				return(myImages)
		return(null)

Action/Movement/StepRight

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		modifier = GetLastDir() // We're not turning 180 here, we're facing the same direction, just turning.
		target = get_step(GetLastStep(),turn(modifier,-90))

		RemoveStepIfInvalid()
		RemoveRedundantMovesIfFound()

	Attempt()

		return(StepMovement())

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		if (target != null)
			var/turf/turfPriorToMove = get_step(target,turn(modifier,90))
			if (turfPriorToMove != null)
				var/imageNewImage = image('imgControlScheme.dmi',turfPriorToMove,"sidestep",3,turn(modifier,-90))
				ShiftImageInItsDirection(imageNewImage)
				myImages.Add(imageNewImage)
				return(myImages)
		return(null)

Action/Movement/StepLeft

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		modifier = GetLastDir() // We're not turning 180 here, we're facing the same direction, just turning.
		target = get_step(GetLastStep(),turn(modifier,90))

		RemoveStepIfInvalid()
		RemoveRedundantMovesIfFound()

	Attempt()

		return(StepMovement())

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		if (target != null)
			var/turf/turfPriorToMove = get_step(target,turn(modifier,-90))
			if (turfPriorToMove != null)
				var/imageNewImage = image('imgControlScheme.dmi',turfPriorToMove,"sidestep",3,turn(modifier,90))
				ShiftImageInItsDirection(imageNewImage)
				myImages.Add(imageNewImage)
				return(myImages)
		return(null)

Action/Movement/TurnRight

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		target = GetLastStep()
		modifier = turn(GetLastDir(),-45)

		RemoveRedundantMovesIfFound()

	Attempt()

		actor.dir = modifier
		actor.ActionTurned()

		return(ACTION_STATUS_COMPLETE)

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		var/turf/turfPriorToMove = target
		if (turfPriorToMove != null)
			myImages.Add(image('imgControlScheme.dmi',turfPriorToMove,"turn",3,modifier))
			return(myImages)
		return(null)

Action/Movement/TurnLeft

	New(var/mob/newActor,var/atom/newTarget,var/newModifier)

		actor = newActor
		target = GetLastStep()
		modifier = turn(GetLastDir(),45)

		RemoveRedundantMovesIfFound()

	Attempt()

		actor.dir = modifier
		actor.ActionTurned()

		return(ACTION_STATUS_COMPLETE)

	GetClientScreenImages()
		if (myImages != null) return(myImages)
		myImages = new/list()
		var/turf/turfPriorToMove = target
		if (turfPriorToMove != null)
			myImages.Add(image('imgControlScheme.dmi',turfPriorToMove,"turn",3,modifier))
			return(myImages)
		return(null)