/*
==================================================
	Project Xenoverse
	By Geldon Yetichsky
	Using the most excellent BYOND suite.
==================================================
actionQueueMob.dm
	Things which should be attached to mob objects for the ActionQueue to work.

==================================================

Notes:
*/

mob

	// Note that, under the current implementation, a action list being set to null is the death flag.
	var/Action/actions[]
	var/Interruption/interruptions[]

	/* Live()
		Initializes the action queue and does any other maintenance involved in allowing the mob to live.
	*/
	proc/Live()

		if (istype(src,/mob/player))
			if (src in respawnArea)
				DebugMsg(".Live() rejected because /mob/player is in respawn area.")
				return()

		if (actions == null) actions = new/list()
		if (interruptions == null) interruptions = new/list()

		spawn()
			ActionHandlerProcess(src)

		density = 1

	/* Die()
		Kills the action queue and does any other maintenance involved in allowing the mob to die.
	*/
	proc/Die()
		 // The reason why I do this before del actions/interruptions is to assure client.screen is cleaned up on death.
		ActionHandlerProcessPurgeAll()
		del actions
		del interruptions

		/* Any special treatment for individual mobs can go on the individual mobs.
		if (istype(src,/mob/player))
			src.Move(locate(/area/respawn))
		else // Leave the corpse for other players to see, but do not allow it to obstruct them
			density = 0
		*/

		// Player or NPC, if you're dead, you don't obstruct people.
		density = 0

	proc/QueueNewAction(var/Action/thisAction)

			/* This is actually normal in the case that the player attempted to perform an impossible move.
			   The Action.New() verifies this and, if it's bad, removes itself from the queue and deletes itself automatically.
			   However, the null result is sent regardless. */

		if (thisAction == null)
			return()

		if (!istype(thisAction,/Action))
			DebugMsg("[type].QueueNewAction() only takes Actions.")
			return()

		if (src.IsAlive())
			actions.Add(thisAction)
			// Uncomment this to verify actions are quing.
			// DebugMsg("QueueNewAction: new [thisAction.type] Action queued.")

		else if (src.client != null)
			src << "You cannot queue a new action: you are dead."

/*  ActionHandlerProcess()
	The ActionHandlerProcess() should be executed on any mob who is expected to sequence their actions using the action queue.
	ActionHandlerProcess() handles these things:

		1. The action queue (mob.actions) - These are things which the mob has sequenced to do.
		2. The interruption queue (mob.interruptions) - These are things which must happen before any actions do.
		3. The resolution of turns (e.g. replenishment of action points).

	When something actually needs to be done (the "action") it is passed off to the appropriate ActionHandlerProcess.  These procedures
	are expected to attempt to perform the action.  They will then return one of these results:

	0 - Cannot do this in the current sequence of actions and/or conditions.
	1 - Can do this or am doing this, but we're not done yet.  Popular result when there's insufficient AP to perform the action.
	2 - This is done, move on to next action.*/

	proc/ActionHandlerProcess()

			// Might as well put these checks here - it's more robust, and exists outside the loop.
		if (actions == null)
			actions = new/list()

		if (interruptions == null)
			interruptions = new/list()

		var/mob/player/thisPlayer

		if (istype(src,/mob/player))	// Since I call this so often, might as well only evaluate it once.
			thisPlayer = src

		while (actions != null) // actions == null is the new alive == 0

			// DebugMsg("ActionHandlerProcess executing.  Actions: [actions.len]")

			if (thisPlayer != null)
				thisPlayer.ExecuteSuitGeneratorIteration() // This might not work if ActionHandlerProcess is basically only run when actions are queued.
				if (thisPlayer.client != null)
					thisPlayer.client.UpdateDisplayPerTurn()

			// Interruption queue handling.  Interruptions are resolved before any actions take place.

			if (InterruptionHandler(src))
				continue

			// Action queue handling.

			if (actions.len == 0)
				sleep(TIMEPERACTION)
				// This is a good place to set a neutral animation.
				// Note checks to see if this is redundant are on the Animation proc.
				continue // This continue is kinda important - without, the next lines try to pull the first action and fails..

			// ActionPoint cost calculcation and deduction.
			var/tmp/Action/currentAction = actions[1]
			var/resultCode = currentAction.Attempt()

			if (currentAction.CannotAfford())
				sleep(TIMEPERACTION)
				continue

			if (resultCode == null)
				DebugMsg("[type].ActionHandlerProcess() null result code on [currentAction].Attempt()")
				ActionHandlerProcessPurgeAll()
				continue

			switch (resultCode)
				if (ACTION_STATUS_IMPOSSIBLE) // The action encountered a critical error state that comprimises the queue.
					ActionHandlerProcessPurgeAll()
					continue
				if (ACTION_STATUS_PENDING) // The action is being done, but is not complete yet.
					DebugMsg("ActionHandlerProcess: Action pending [currentAction.type]")
					continue
				if (ACTION_STATUS_COMPLETE) // Consider this action complete and move on to the next.
					ActionHandlerProcessNextAction()
					continue
				else
					DebugMsg("[type].ActionHandlerProcess() unknown result code [resultCode] on [currentAction.type].Attempt()")
					ActionHandlerProcessPurgeAll()
					continue

				/*  This is how abilities used to work, before I made obj/abilities merely spawn Actions.

				if ("UseAbility") // This is what does all the program execution
					sleep(10)
					if (istype(currentAction.what,/obj/ability))
						var/tmp/obj/ability/thisAbility = currentAction.what

						switch (thisAbility.Activate(currentAction))
							if (0) // Program cannot be executed
								ActionHandlerProcessNextAction(thisMob)
								continue
							if (1)
								continue
							if (2)
								ActionHandlerProcessNextAction(thisMob)
								continue
					else // If the currentAction.what is not an /obj/ability
						DebugMsg("ActionQueue() current program failure: Non-program on what of current action.")
						ActionHandlerProcessNextAction(thisMob)
						continue

				else
					DebugMsg("ActionQueue() Unrecognized Action: [currentAction.action]")
					ActionHandlerProcessNextAction(thisMob)
					continue
				*/

			if (actions != null)
				DebugMsg("ActionQueue() loose flow: [src]: [currentAction.type]")
				continue

		// If the flow falls down here, thisMob.actions == null.

	proc/ActionHandlerProcessNextAction()

		if (actions == null)
			// That just means you're dead is all.
			// DebugMsg("Null action list on [name]")
			return()

		if (actions.len > 0)
			if (client != null)
				client.RemoveClientScreenAction(actions[1])
			actions.Cut(1,2)
		else
			ActionHandlerProcessPurgeAll()

		/* Purges everything from action queue. */
	proc/ActionHandlerProcessPurgeAll()
		if (client != null)
			client.RemoveClientScreenActionAll()
		del actions
		actions = new/list()


	proc/ActionHandlerProcessRemoveLast()
		if (actions.len > 0)
			var/Action/actionToRemove = actions[actions.len]
			actions.Remove(actionToRemove)
		else
			return()

	/*
		DrawLatestAction()
		Just a stopgap measure to draw actions to client.screen as soon as they're added... probably.  Basically, just
		redraws the last action in the queue.  Should be called from wherever... probably immediately after triggering the
		obj/ability in question.
	*/
	proc/DrawLatestAction()
		if (client != null)
			if (actions != null && actions.len > 0)
				client.AddClientScreenAction(actions[actions.len])

	proc/RemoveActionFromQueue(var/Action/actionToRemove)

		for (var/Action/thisAction in actions)
			if (thisAction == actionToRemove)
				if (client != null)
					client.RemoveClientScreenAction(actionToRemove)
				actions.Remove(thisAction)

	/* NewAction()

		Now deprecated by using action type to resolve the type of action.

		Old Notes:

	   This is basically a quick a robust way to add a new action to the queue.
	   Arg 1, the action, is required.  It's a string indicating what's to be done, e.g. "move" or "get"
	   Arg 2, the subject, is not required for some actions.  It's what it is being done to.  e.g. a turf to move to.
	   Arg 3, the what, is a potential modifier to be added to the action.  e.g. An item being used.
	   I'd pass an action if I could, but basically this is where the actions >are created< so that's kinda impossible.
	*/

	/* Actually, this isn't doing a whole lot I'm not already doing.

	proc/NewAction(var/newActionType,var/newActionSubject,var/newActionWhat)

		if (newActionType == null)
			DebugMsg("mob.NewAction(): Given action null.")
			return(0)

		if (actions == null) // This means you're dead, you can't make new actions.
			return()
			 // actions = new/list()

		var/newAction = new/newActionType(src,newActionSubject,newActionWhat)

		src.actions.Add(newAction)

		if (newAction.actor.client != null)
			newAction.actor.client.AddClientScreenActionAll()
	*/

	/* NewInterrupt()
	   This is basically a quick a robust way to add a new interruption to the queue.
	   Arg 1, the action, is required.  It's a string indicating what's to be done, e.g. "move" or "get"
	   Arg 2, the subject, is not required for some actions.  It's what it is being done to.  e.g. a turf to move to.
	   Arg 3, the what, is a potential modifier to be added to the action.  e.g. An item being used.
	   I'd pass an action if I could, but basically this is where the actions >are created< so that's kinda impossible.
	*/
	proc/NewInterrupt(var/newActionName,var/newActionSubject,var/newActionWhat)

		if (newActionName == null)
			DebugMsg("mob.NewAction(): Given action null.")
			return(0)

		if (actions == null) // This means you're dead, you can't make new interruptions.
			return()

		var/Interruption/newAction = new/Action()
		newAction.action = newActionName

			// Now that the new action has been created, time to assign values to it.

		newAction.actor = src // The actor, useful to have on an Action variable that gets passed to who knows where.

		if (newActionSubject != null) // The thing it is being done to.  Not necessarily existent
			newAction.subject = newActionSubject

		if (newActionWhat != null) // Modifier of the action.
			newAction.what = newActionWhat

			// All done, action is added.

		src.interruptions.Add(newAction)

		if (newAction.actor.client != null)
			newAction.actor.client.AddClientScreenActionAll()

	/* GetActionString can now, in proper Object Oriented form, be located on the individual actions.
	proc/GetActionString(var/actionNumber)
		if (actionNumber == null || actionNumber > actions.len)
			return("None")
		else
			var/tmp/Action/currentAction = actions[actionNumber]
			switch (currentAction.action)
				if ("move")
					var/tmp/turf/turfMovedTo = currentAction.subject
					return("'Moving to [turfMovedTo.x],[turfMovedTo.y],[turfMovedTo.z].'")
				else
					return("'[currentAction.action]'") */

	/* CancelMoves()
		mob.CancelMoves() simply adds a move purge to the interruption queue. */
	proc/CancelMoves()
		var/tmp/Interruption/thisInterruption = new/Interruption()
		thisInterruption.actor = src
		thisInterruption.action = "Purge"
		thisInterruption.what = null
		thisInterruption.subject = null

		src.interruptions.Add(thisInterruption)