/*
	This is basically all the code related to adjusting the 3 big commodities in the game: energy, mass, and nutrition
	They're pretty close to identical calls, so I set them up in 3 big congruent piles.
	No code duplication is done for efficiency's sake.

	Procs include:
	AdjustCommodity - Shifts the commodity by the value listed here, positive or negative, does not allow it to go under 0 or
		higher than GetMaxCommodity.

	GetCommodity - Gets the current value of the commodity.

	SetCommodity - Simply sets the commodity to this value.  Does not allow it to go higher than maximum.  Unlike
	    AdjustCommodity it will generate an error messaage if this happens, because chances are it's being utilized by a
	    process which assumes it can be done.

	GetMaxCommodity - Gets the maximum value allowed for that commodity

	SetMaxCommodity - Sets the maximum value allowed for that commodity.

	All procs assign default values to commodities or maximum values if they are found to be null.

	AdjustCommodity, SetCommodity, and SetMaxCommodity will also update the client skin if it is not null.

	Commodities are allowed to exist at negative numbers, indicating a shortage, but this is sort of illogical.
*/

mob

	/*
		This is called on mob.New() - it's primary purpose is to obliterate null values.
		It sets the values directly rather than using the appropriate procs so as not to update the client skin
		prematurely.
	*/
	proc/DefaultNewMobCommoditySetup()
		if (mass == null) mass = DEFAULT_MOB_MASS
		if (maxMass == null) maxMass = DEFAULT_MOB_MASS_MAX
		if (energy == null) energy = DEFAULT_MOB_ENERGY
		if (maxEnergy == null) maxEnergy = DEFAULT_MOB_ENERGY_MAX
		if (nutrition == null) nutrition = DEFAULT_MOB_NUTRITION
		if (maxNutrition == null) maxNutrition = DEFAULT_MOB_NUTRITION_MAX

	/***************
		Energy
	****************/

var/DEFAULT_MOB_ENERGY = 0
var/DEFAULT_MOB_ENERGY_MAX = 100

mob

	var/energy // Energy count.  Used to perform
	var/maxEnergy // Energy cannot exceed this amount.

	proc/AdjustEnergy(var/amountToAdjustBy)
		if (!(isnum(amountToAdjustBy)))
			DebugMsg("[type].AdjustEnergy() passed non-numeric argument: [amountToAdjustBy]")
		else
			energy = GetEnergy() + amountToAdjustBy
			if (energy > GetMaxEnergy())
				energy = GetMaxEnergy()

		if (client != null)
			client.SkinUpdateEnergy()

	proc/SetEnergy(var/newValue)
		if (isnum(newValue))
			if (newValue > GetMaxEnergy())
				DebugMsg("[type].SetEnergy() cannot set energy higher than maximum.")
				energy = GetMaxEnergy()
			else
				energy = newValue
		else
			DebugMsg("SetEnergy() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateEnergy()

	proc/GetEnergy()
		if (energy == null) energy = DEFAULT_MOB_ENERGY
		return (energy)

	proc/SetMaxEnergy(var/newValue)
		if (isnum(newValue))
			maxEnergy = newValue
			if (energy > maxEnergy)
				energy = maxEnergy
				DebugMsg("[type].SetMaxEnergy() reduced current value to match on [name].")
		else
			DebugMsg("SetMaxEnergy() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateEnergy()

	proc/GetMaxEnergy()
		if (maxEnergy == null) maxEnergy = max(GetEnergy(),DEFAULT_MOB_ENERGY_MAX)
		return (maxEnergy)

	/***************
		Mass
	****************/

var/DEFAULT_MOB_MASS = 0
var/DEFAULT_MOB_MASS_MAX = 100

mob

	var/mass // Mass count.  Used as armor as to perform mass fabrication.
	var/maxMass // Mass should not exceed this amount

	proc/AdjustMass(var/amountToAdjustBy)
		if (!(isnum(amountToAdjustBy)))
			DebugMsg("[type].AdjustMass() passed non-numeric argument: [amountToAdjustBy]")
		else
			mass = GetMass() + amountToAdjustBy
			if (mass > GetMaxMass())
				mass = GetMaxMass()

		if (client != null)
			client.SkinUpdateMass()

	proc/SetMass(var/newValue)
		if (isnum(newValue))
			if (newValue > GetMaxMass())
				DebugMsg("[type].SetMass() cannot set energy higher than maximum.")
				mass = GetMaxMass()
			else
				mass = newValue

		else
			DebugMsg("SetMass() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateMass()

	proc/GetMass()
		if (mass == null) mass = DEFAULT_MOB_MASS
		return (mass)

	proc/SetMaxMass(var/newValue)
		if (isnum(newValue))
			maxMass = newValue
			if (mass > maxMass)
				mass = maxMass
				DebugMsg("[type].SetMaxMass() reduced current value to match on [name].")
		else
			DebugMsg("SetMaxMass() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateMass()

	proc/GetMaxMass()
		if (maxMass == null) maxMass = max(GetMass(),DEFAULT_MOB_MASS_MAX)
		return (maxMass)

	/***************
		Nutrition
	****************/

var/DEFAULT_MOB_NUTRITION = 0
var/DEFAULT_MOB_NUTRITION_MAX = 100

mob

	var/nutrition // Nutrition count.  Living creatures need it to survive.
	var/maxNutrition // Nutrition should not exceed this amount.

	proc/AdjustNutrition(var/amountToAdjustBy)
		if (!(isnum(amountToAdjustBy)))
			DebugMsg("[type].AdjustNutrition() passed non-numeric argument: [amountToAdjustBy]")
		else
			nutrition = GetNutrition() + amountToAdjustBy
			if (nutrition > GetMaxNutrition())
				nutrition = GetMaxNutrition()

		if (client != null)
			client.SkinUpdateNutrition()

	proc/SetNutrition(var/newValue)
		if (isnum(newValue))
			if (newValue > GetMaxNutrition())
				DebugMsg("[type].SetNutrition() cannot set energy higher than maximum.")
				nutrition = GetMaxNutrition()
			else
				nutrition = newValue
		else
			DebugMsg("SetNutrition() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateNutrition()

	proc/GetNutrition()
		if (nutrition == null) nutrition = DEFAULT_MOB_NUTRITION
		return(nutrition)

	proc/SetMaxNutrition(var/newValue)
		if (isnum(newValue))
			maxNutrition = newValue
			if (nutrition > maxNutrition)
				nutrition = maxNutrition
				DebugMsg("[type].SetMaxNutrition() reduced current value to match on [name].")
		else
			DebugMsg("SetMaxNutrition() requires a numeric arg.")
			return()

		if (client != null)
			client.SkinUpdateNutrition()

	proc/GetMaxNutrition()
		if (maxNutrition == null) maxNutrition = max(GetNutrition(),DEFAULT_MOB_NUTRITION_MAX)
		return(maxNutrition)

