/*
	Projectiles are central to this project.  After doing several projectiles, I realized that they all have the same general
	behavior except they differ in their payload and whether or not they can pass walls.  This lead to the realization that
	this would be an ideal object oriented programming exercise.
*/

// Projectile speeds are measured in MS delay (of course)
var/const/DEFAULT_PROJECTILE_SPEED = 2 // 2 == 5 spaces/sec
var/const/DEFAULT_PROJECTILE_POWER_LOSS = 1 // Amount of power a projectile loses when not impacting something by default.

obj/projectile

	icon = 'imgProjectiles.dmi'

	mouse_opacity = 0 // Projectiles cannot be clicked, merely witnessed.
	density = 0 // Projectiles can run into things... but are not in themselves considered something to run into.

		/* Range reflects the maximum number of spaces that a projectile can travel before it expires.  If not defined
		  it is assumed the projectile will keep going until it runs out of energy or hits a barrier */
	var/range

		/* Power is closely related to Attack.power.  It tends to diminish as the projectile travels, with exceptions. */
	var/power

		/* When it comes to firing a projectile, we definately need to know where it's going, that's where the source and
		  destination variables come in.  Note that projectiles in this game have a life of their own: the firers
		  shoot them in a given direction, but if the projectile has reached the destination and is not set to have a
		  limited range then the projectile will simply keep going until it runs out of power or hits something. */
	var/atom/source // The place that the projectile was originally fired from.

	var/atom/destination // The place that the projectile was originaly fired at.

	var/turf/path[] // Each projectile has a calculated path.

		/* It's the intent of this game that it will keep track of which players damage what.  The responsibleParty variable
		   stores this information. */
	var/mob/player/responsibleAtom

		/* There's one more very important consideration for projectiles: the /type/ of projectile being fired.  This is
		  actually tracked via it's .type as different Projectiles will have different mechanisms which react differently
		  to things encountered along the way. */

	/*
		Given the nature of projectiles, one New() proc can handle all types, and differences in behavior can be evaluated
		based on type.
	*/

	New (var/atom/newSource,var/atom/newDestination,var/newPower,var/newRange,var/atom/newResponsibleAtom)

		if (newSource == null)
			DebugMsg("[type].New() projectile could not be generated as it has no source.")
			del src
			return()
		else
			source = newSource

		if (newDestination == null)
			DebugMsg("[type].New() projectile could not be generated as it has no destination.")
			del src
			return()
		else
			destination = newDestination

		if (newPower == null)
			DebugMsg("[type].New() projectile could not be generated as it has no power.")
			del src
			return()
		else
			power = newPower

		if (newRange != null)
			if (isnum(newRange))
				range = newRange
			else
				DebugMsg("[type].New() 4th arg was not a number or null.")
				// Keep running regardless.

		if (newResponsibleAtom != null)
			if (istype(newResponsibleAtom,/atom))
				responsibleAtom = newResponsibleAtom
			else
				DebugMsg("[type].New() 5th arg was not an atom.")
				// Keep running regardless.

		dir = get_dir (source,destination)

		del path
		path = new/list()

		path = LineOfFire(source,destination,10,2) // Calculate 10 squares at a time.

		spawn() Travel()

	/*
		The Travel() proc is spawned() from New() to simulate the passage of the projectile.
	*/
	proc/Travel()

		if (path == null)
			DebugMsg("[type].path was null!")
			del src
			return()

		while (power > 0 || range == 0)

				// First thing's first: the projectile travels to the next step on the path.

			if (path == null || path[1] == null) // If it has no path, something went wrong, terminating it here.
				power = 0
				continue
			else
				Move(path[1])
				/* The range a projectile travels is decreased each space.  When it reaches 0, the projectile terminates.
				   after going through this last iteration. */
				if (range != null && range != 0) range--

				// First concern: cosmetic FX per round.  How superficial of me.

			UpdateAppearance()

				// Second concern: doing stuff to the things what deserve it that may reside in this hex

			for (var/mob/thisMob in loc)
				Impact(thisMob)

			if (power <= 0) continue

			for (var/obj/device/thisDevice in loc)
				Impact(thisDevice)

			if (power <= 0) continue

				/* Third concern: behaviors related to how this projectile behaves in the given space.
					Note: this is a choice to have the things that reside within a space impacted before the space
					itself.  It's easily enough removed by swapping this code. */

			var/turf/thisTurf = src.loc
			TurfAffect(thisTurf)

			if (power <= 0) continue

				// Now that we have moved, we make sure we have a place to move next turn.  It will recalculate it if the path is spent.

			if (path.len != 0)

				if (path.len > 2)
					// Here, I am basically appending to the path as I go.
					var/directionFlow = get_dir(path[1],path[2])
					var/turf/newTurf = get_step(path[path.len],directionFlow)
					if (newTurf != null)
						path.Add(newTurf)

				path.Cut(1,2)

			if (path.len == 0) // This should only happen if for some reason it can't append ot the path.
				RecalculatePathBasedOnDirection()

			if (path == null || path.len == 0) // This should only happen if it encountered difficulty calculating it, probably off-map.
				if (x != 0 && x != world.maxx && y != 0 && y != world.maxy) // This is normal if it hit world edge
					DebugMsg("[type] stopped at [x],[y],[z] because it was unable to path.")
				power = 0
			else
				sleep(GetProjectileSpeed()) // 5 hexes a second.

		/* Now that we're done traveling, there's a question: did it stop traveling because it ran out of range, or did it
			stop traveling if it ran out of power?  For the time being, it doesn't matter, but in the future perhaps I'll
			have a special call for that. */

		TerminateProjectile()

	/* RecalculatePathBasedOnDirection
		Warning! This should not be called indescriminantly because it will basically delete and redraw the path based off of
		the current direction of the projectile.  Why this is problematic is because, by default, the path can go in more
		than just the 8 compass directions.  It does this by appending the last direction onto the end every time one is
		cut. */

	proc/RecalculatePathBasedOnDirection()
		del path
		path = new/list()
		if (range != null && range < 10 && range > 1)
			path.Add(LineOfFire(loc,get_step(loc,dir),range,2)) // Calculate only as many squares as are needed
		else
			path.Add(LineOfFire(loc,get_step(loc,dir),10,2)) // Calculate 10 squares at a time.

	proc/GetProjectileSpeed()
		return(DEFAULT_PROJECTILE_SPEED)

	/*
		Appearance changes depending on the type of projectile.  If it has fallen to root level then this
		projectile does not have an appearance.
	*/
	UpdateAppearance()
		// DebugMsg("[type].UpdateAppearance() fell to parent level.")
		..()

	/*
		Different projectiles do differnt things when they impact them.  A root level projectile doesn't have
		an effect so it's an error if it falls down here.
	*/
	proc/Impact(var/atom/impactedThis)
		DebugMsg("[type].Impact() fell to the parent level.")

	/*
		Default TurfAffect() behavior is to have the projectile attack the turf if it has density (which is BYOND's way
		of indicating a wall) or otherwise ignore it.
	*/
	proc/TurfAffect(var/turf/thisTurf)
		if (thisTurf.density)
			var/Attack/newAttack = new(power)
			thisTurf.Damage(newAttack)
			power = 0
		else
			power = power - DEFAULT_PROJECTILE_POWER_LOSS

	/*
		Default TerminateProjectile() behavior is to simply delete the projectile, but some projectiles might have a
		special effect that mandates delaying their deletion.
	*/
	proc/TerminateProjectile()
		// DebugMsg("Projectile terminated: path.len [path.len], power [power] range [range]")
		del src

/*
	/obj/projectile/energy
	Energy projectiles exhibit unique behavior when they encounter a solid wall.
	Depending on the material that makes up that wall, the energy projectile may be absorbed or change direction entirely!
	Furthermore, energy projectiles power lost will vary depending on what they hit.
*/

var/const/ENERGY_PROJECTILE_COST_FOR_SOIL_WALL = 1
var/const/ENERGY_PROJECTILE_COST_FOR_SOIL_FLOOR = 1
var/const/ENERGY_PROJECTILE_COST_FOR_PLEXWEAVE_WALL = 100
var/const/ENERGY_PROJECTILE_COST_FOR_PLEXWEAVE_FLOOR = 1
var/const/ENERGY_PROJECTILE_COST_FOR_ALCHELITE_WALL = 0
var/const/ENERGY_PROJECTILE_COST_FOR_ALCHELITE_FLOOR = 0
var/const/ENERGY_PROJECTILE_COST_FOR_SUBNEUTRONIUM_WALL = 2
var/const/ENERGY_PROJECTILE_COST_FOR_SUBNEUTRONIUM_FLOOR = 2

var/const/ENERGY_PROJECTILE_MAX_WEAK_BOLT_POWER = 10
var/const/ENERGY_PROJECTILE_MAX_MEDIUM_BOLT_POWER = 25
var/const/ENERGY_PROJECTILE_MAX_HIGH_BOLT_POWER = 50

// Recommended I keep these relatively high, it can be really hard to hear them.
var/const/ENERGY_PROJECTILE_VOLUME_LOW_INTENSITY = 80
var/const/ENERGY_PROJECTILE_VOLUME_MEDIUM_INTENSITY = 90
var/const/ENERGY_PROJECTILE_VOLUME_HIGH_INTENSITY = 95
var/const/ENERGY_PROJECTILE_VOLUME_EVEN_HIGHER_INTENSITY = 100

obj/projectile/energy
	density = 0

	icon = 'imgProjectiles.dmi'

	luminosity = 2 // That's right, ball lightning casts light.

	UpdateAppearance()
		switch (power)
			if (0 to ENERGY_PROJECTILE_MAX_WEAK_BOLT_POWER) // Weak bolt
				Noise(src,'electricbolt.wav',ENERGY_PROJECTILE_VOLUME_LOW_INTENSITY)
				icon_state = "weakbolt"

			if (ENERGY_PROJECTILE_MAX_WEAK_BOLT_POWER to ENERGY_PROJECTILE_MAX_MEDIUM_BOLT_POWER) // Normal bolt
				Noise(src,'electricbolt2.wav',ENERGY_PROJECTILE_VOLUME_MEDIUM_INTENSITY)
				icon_state = "normalbolt"

			if (ENERGY_PROJECTILE_MAX_MEDIUM_BOLT_POWER to ENERGY_PROJECTILE_MAX_HIGH_BOLT_POWER) // Strong bolt
				Noise(src,'electricbolt3.wav',ENERGY_PROJECTILE_VOLUME_HIGH_INTENSITY)
				icon_state = "strongbolt"

			else // Maximum power bolt
				Noise(src,'electricbolt4.wav',ENERGY_PROJECTILE_VOLUME_EVEN_HIGHER_INTENSITY)
				icon_state = "maximumbolt"


	Impact(var/atom/impactedThis)

		if (istype(impactedThis,/mob))
			var/mob/thisMob = impactedThis
			thisMob.Jolt(power)
			power = 0

		else if (istype(impactedThis,/obj/device))
			var/obj/device/thisDevice = impactedThis
			var/maxAbsorbed = thisDevice.GetMaxEnergy() - thisDevice.GetEnergy()
			if (maxAbsorbed != 0)
				if (maxAbsorbed >= power)
					// DebugMsg("Impact.Jolt([power])")
					thisDevice.Jolt(power)
					power = 0
				else
					thisDevice.Jolt(maxAbsorbed)
					power = power - maxAbsorbed

			/* Bah, this is too bloaty and too much work: lets just use Jolt.
			var/maximumEnergyAbsorbed = (impactedThis.GetMaxEnergy() - impactedThis.GetEnergy())
			var/amountOfEnergyAbsorbed = max(maximumEnergyAbsorbed,power)
			var/amountToAttackWith
				// If the amount of energy it can absorb matches or exceeds the amount of power that impacted it
			if (maximumEnergyAbsorbed >= power)
				// It absorbs all the power
				amountOfEnergyAbsorbed = power
				// And takes no damage
				amountToAttackWith = 0
			else
				// Otherwise, it is only fried by the amount of energy it didn't absorb.
				amountToAttackWith = power - amountOfEnergyAbsorbed

			if (amountOfEnergyAbsorbed > 0)
				thisMob.Jolt(amountOfEnergyAbsorbed)
			if (amountToAttackWith > 0)
				var/Attack/newAttack = new(amountToAttackWith,ELECTRICAL_ATTACK_TYPE)
				if (responsibleAtom != null)
					newAttack.responsibleAtom = responsibleAtom
				thisMob.Damage(newAttack)

			power = power - (amountOfEnergyAbsorbed + amountToAttackWith)
			if (power < 0) power = 0
			*/

	TurfAffect(var/turf/thisTurf)

		switch (thisTurf.GetDensity())
			if (0)
				if (istype(thisTurf,/turf/wall)) // Soil walls
					power = power - ENERGY_PROJECTILE_COST_FOR_SOIL_WALL
				else							// Soil floors
					power = power - ENERGY_PROJECTILE_COST_FOR_SOIL_FLOOR
			if (1)
				if (istype(thisTurf,/turf/wall)) // Plexweave walls
					power = power - ENERGY_PROJECTILE_COST_FOR_PLEXWEAVE_WALL
				else							// Plexweave floors
					power = power - ENERGY_PROJECTILE_COST_FOR_PLEXWEAVE_FLOOR

			if (2)		// Alchelite is capable of catching and redirecting energy.
				if (istype(thisTurf,/turf/floor))
					power = power - ENERGY_PROJECTILE_COST_FOR_ALCHELITE_FLOOR
				else // Alchelite walls
					power = power - ENERGY_PROJECTILE_COST_FOR_ALCHELITE_WALL

					// These walls attempt to redirect the current to another nearby  Alchelite wall.
					var/tmp/list/turf/theseTurfs = new/list()
					for (var/tmp/turf/findAlcheliteTurf in cardinal(src))
						if (findAlcheliteTurf != null && istype(findAlcheliteTurf,/turf/wall/alchelite))
							theseTurfs.Add(findAlcheliteTurf)

					var/tmp/reverseTurf = get_step(loc,turn(dir,180)) // unless it's backwards.
					if (reverseTurf in theseTurfs)
						theseTurfs.Remove(reverseTurf)

					if (theseTurfs.len > 0) // Change path to another alchelite turf..
						var/tmp/sameDirTurf = get_step(loc,dir) // same direction preferred
						if (!(sameDirTurf in theseTurfs))
							dir = get_dir(loc,theseTurfs[rand(1,theseTurfs.len)])
							del path
							RecalculatePathBasedOnDirection()
						// else // SameDirTurf *is* in the same direction.
					// else // There was no turfs to redirect to.

			if (3)
				if (istype(thisTurf,/turf/wall))	// Subneutronium wall
					power--		// Being pretty darn solid, absorb two levels of power
					power--
				else		// Subneutronium floor
					power = power	// No impact - flew right over
			else // Transmatrix or special wall.
				if (istype(thisTurf,/turf/wall))
					if (istype(thisTurf,/turf/wall/transmatrix/transmatrixtransparent))
						power = power // Specially formulated to let energy pass through unmodified
					else
						power = 0
				else								// Transmatrix or nondescript floor
					power = power   // No impact - flew right over.

			// Creation of the groundsparks effect.  Always appears on solid walls, otherwise a 1 in 4 chance of appearing
		if (loc.density)
			new/obj/effects/groundsparks(src.loc,turn(dir,180)) // Passing my own dir as a variable, I hope that works.
		else if (rand(1,4) == 4)
			new/obj/effects/groundsparks(src.loc,dir)


/*  /obj/projectile/mass
	This has evolved into a sort of a universal constructor facilitator.  If it impacts damage, it repairs it.  When it reaches
	the end of its path, it does not extend the path, instead it tries to build some turf there.

	I still haven't decided what happens when it impacts something other than terrain.  Maybe it'll do a mass attack, or
	maybe something more interesting. */

obj/projectile/mass

	icon = 'imgProjectiles.dmi'

	luminosity = 1 // Mass projectiles cast a little bit of light.  Oh, those fancy nanites.

	var/const/MASS_PROJECTILE_POWER_LOSS = 2 // How much power is lost per hex traveled as a mass projectile.
	var/const/POWER_REQUIRED_TO_TRANSMUTE_SOIL = 0
	var/const/POWER_REQUIRED_TO_TRANSMUTE_PLEXWEAVE = 4
	var/const/POWER_REQUIRED_TO_TRANSMUTE_ALCHELITE = 8
	var/const/POWER_REQUIRED_TO_TRANSMUTE_SUBNEUTRONIUM = 16

	UpdateAppearance()
		// The reason why I'm taking a note here is I anticipate I won't be using mass bolts.
		DebugMsg("/obj/projectile/mass.UpdateAppearance() needs updating.")
		switch(power)
			if (1) // Weak mass
				Noise(src,'electricbolt.wav',3) // Should eventually make this sound uniquely like mass.
				icon_state = "weakmass"

			if (2) // Normal mass
				Noise(src,'electricbolt2.wav',5)
				icon_state = "normalmass"

			if (3) // Strong mass
				Noise(src,'electricbolt3.wav',7)
				icon_state = "strongmass"

			// Again, I should really consider adding a 4th kind of mass bolt effect.

			else // Maximum mass
				Noise(src,'electricbolt4.wav',10)
				icon_state = "maximummass"

	/////////////////////////////////////

	// Doesn't do anything to critters or devices... yet

	Impact(var/atom/impactedThis)
		return()

	TurfAffect(var/turf/thisTurf)

			// Mass bolts are capable of repairing damaged turfs.
		for (var/obj/turfdamage/thisDamage in thisTurf)
			var/amountOfDamageRepaired = min(power,thisDamage.damageSustained)
			power = power - amountOfDamageRepaired
			thisDamage.Repair(amountOfDamageRepaired)
			break

		if (power == 0)
			return()

		// Mass bolts that have reached their destination will attempt to construct a new turf there.

		/* I get the feeling I'll be phasing this behavior out.  It's a crude construction tool at best.  Instead,
		  seperate abilities that increase/decrease a turf's density or raise/lower them (walls/floors) would be ideal.
		  It has a major world-muting effect, which is another reason I have to rethink this */

		if (istype(thisTurf,/turf/floor))

			switch (thisTurf.GetDensity())
				if (0) // Affecting soil
					if (power > POWER_REQUIRED_TO_TRANSMUTE_SOIL) // Moot?
						new/turf/floor/plexweave(loc)
					power = 0

				if (1) // Affecting plexweave floor
					if (power > POWER_REQUIRED_TO_TRANSMUTE_PLEXWEAVE)
						new/turf/floor/alchelite(loc)
					power = 0

				if (2) // Affecting alchelite floor
					if (power > POWER_REQUIRED_TO_TRANSMUTE_ALCHELITE)
						new/turf/floor/subneutronium(loc)
					power = 0

				if (3) // Affecting subneutronium floor.  This is the only way players can create walls.
					if (power > POWER_REQUIRED_TO_TRANSMUTE_SUBNEUTRONIUM)
						// Because the player is creating a wall, it's necessary to call the wall drawing function.
						new/turf/wall/subneutronium(loc)
					power = 0

				if (4) // Can't affect transmatrix
					power = 0

		else if (istype(thisTurf,/turf/wall))
			switch (thisTurf.GetDensity())
				if (0) // Affecting soil
					if (power > POWER_REQUIRED_TO_TRANSMUTE_SOIL) // Moot?
						new/turf/wall/plexweave(loc)
					power = 0

				if (1) // Affecting plexweave wall
					if (power > POWER_REQUIRED_TO_TRANSMUTE_PLEXWEAVE)
						new/turf/wall/alchelite(loc)
					power = 0

				if (2) // Affecting alchelite wall
					if (power > POWER_REQUIRED_TO_TRANSMUTE_ALCHELITE)
						new/turf/wall/subneutronium(loc)
					power = 0

				else // Can't affect transmatrix, can't upgrade subneutronium any further than it already is
					power = 0

		power = power - MASS_PROJECTILE_POWER_LOSS

		if (power < 0) power = 0

/*	obj/projectile/force
	Force projectiles basically just slam something with a raw wall of force (accelerated and compacted air, perhaps).
 */

obj/projectile/force

	icon = 'imgProjectiles.dmi'
	icon_state = "force"

	var/const/FORCE_PROJECTILE_POWER_LOSS = 2

	TurfAffect(var/turf/thisTurf)

		if (thisTurf.density)
			var/Attack/thisAttack = new()
			thisAttack.power = power
			// DebugMsg("[type] damaging foe [thisMob.name]") // Important to put this before the damage, else mob may be del'd
			thisTurf.Damage(thisAttack)
			power = 0
			return()

		power = power - FORCE_PROJECTILE_POWER_LOSS

		if (power < 0) power = 0

		if (power > 0) // Which means we'll survive to the next stage.
			// Force projectiles leave behind a little FX in their wake
			new/obj/effects/railslice(loc,dir)


	Impact(var/atom/impactedThis)
		if (istype(impactedThis,/mob))
			var/mob/thisMob = impactedThis
			var/Attack/thisAttack = new()
			thisAttack.power = power
			// DebugMsg("[type] damaging foe [thisMob.name]") // Important to put this before the damage, else mob may be del'd
			thisMob.Damage(thisAttack)
			power = 0

/*  obj/projectile/fletchette
	These are basically simple machine-gun like attacks.  They move so fast that, to the player's perceptions,
	they are instant.  The bolt is not visible, you just see the impact. */

/obj/projectile/fletchette

	icon = 'imgProjectiles.dmi'
	icon_state = "" // Invisible until impact.

	layer = 5 // Fletchette impact is on the top layer so it's always visible.

	var/const/FLETCHETTE_POWER_LOST_PER_SPACE = 1

	TurfAffect(var/turf/thisTurf)

		if (thisTurf.density) // Hit a wall.
			var/Attack/thisAttack = new()
			thisAttack.power = power
			thisTurf.Damage(thisAttack)
			power = 0


	Impact(var/atom/impactedThis)
		if (istype(impactedThis,/mob))
			var/mob/thisMob = impactedThis
			var/Attack/thisAttack = new()
			thisAttack.power = power
			// DebugMsg("[type] damaging foe [thisMob.name]") // Important to put this before the damage, else mob may be del'd
			thisMob.Damage(thisAttack)
			power = 0
			return()
		if (istype(impactedThis,/obj/device))

			if (istype(impactedThis,/obj/device/shieldgen))
				var/obj/device/shieldgen/thisShieldGen = impactedThis
				if (thisShieldGen.IsAccessPermittedFor(src))
					return()

			var/obj/device/thisDevice = impactedThis
			var/Attack/thisAttack = new()
			thisAttack.power = power
			// DebugMsg("[type] damaging foe [thisMob.name]") // Important to put this before the damage, else mob may be del'd
			thisDevice.Damage(thisAttack)
			power = 0
			return()

	TerminateProjectile()
		flick ("machinegunrichochets",src)
		sleep(10)

