/*
	Project Xenoverse
	By Geldon Yetichsky

file:
Proc.dm

notes:
Handles all global procs which cannot be better associated with a specific category.

*/

proc/DebugMsg(var/text)
	if (debugText == 1)
		world << text

	/* Moves all mobs and objects associated with being movable objects.  Useful for things such as lifts.
	Takes two turfs. */
proc/MoveAllLooseThings (var/turf/sourceTurf,var/turf/destinationTurf)

	if (sourceTurf == null || destinationTurf == null)
		DebugMsg("MoveAllLooseThings: passed null turfs.")
		return(0)

	if (!(istype(sourceTurf,/turf)) || !(istype(destinationTurf,/turf)))
		DebugMsg("Non/turf passed.  sourceTurf: [sourceTurf].  destinationTurf: [destinationTurf]")

	for (var/mob/thisMob in sourceTurf)
		// DebugMsg("Moving [thisMob] to [destinationTurf]")
		thisMob.loc = destinationTurf

	for (var/obj/thisObject in sourceTurf)
		if (!(istype(thisObject,/obj/effects))) // this is my effect check
			thisObject.loc = destinationTurf

/*
	Noise()
	This is a generic handler for all noises generated on the maps.  It should be made relatively efficient.
	An idea I had that is not implemented at the time of this writing is to make hte following revamps:
	1. check mob.client, if null then skip them.
	2. getline to the mob.  Call sound dampening proc on each turf between the noise and the mob to reduce the
	   volume by just that much.
	3. Nest the noises in a sound object and make the appropriate adjustments to create the illusion of space.
*/

var/const/VOLUME_FALLOFF_PER_SPACE = 5

proc/Noise(var/atom/soundSource,var/soundFile, var/initialVolume)
	var/volumeToPlay

	if (initialVolume == null)
		volumeToPlay = 100
	else
		volumeToPlay = initialVolume

	var/sound/thisSound

	for (var/mob/iHeardThat in range((round(initialVolume/VOLUME_FALLOFF_PER_SPACE)),soundSource))

			// If a tree falls in a forest and nobody is there to hear it...
		if (isnull(iHeardThat.client))
			continue

		thisSound = new/sound(soundFile)
		thisSound.environment = iHeardThat.GetSoundEnvironment()
		thisSound.x = soundSource.x - iHeardThat.x
		thisSound.y = soundSource.y - iHeardThat.y

		thisSound.falloff = 20 // I've got a rather specific mechanism of handling this, which follows

		var/list/turf/turfsBetweenHearerAndSoundSource = getline(iHeardThat,soundSource)

		if (turfsBetweenHearerAndSoundSource != null)
			for (var/turf/thisTurf in turfsBetweenHearerAndSoundSource)
				if (thisTurf == turfsBetweenHearerAndSoundSource[1])
					// We skip the first turf.  If you're in the same turf as the sound, there's no falloff at all.
					continue

				volumeToPlay = volumeToPlay - thisTurf.GetVolumeMuffled()

				volumeToPlay = volumeToPlay - VOLUME_FALLOFF_PER_SPACE
				if (volumeToPlay <= 0)
					return()

		thisSound.volume = volumeToPlay
		thisSound.channel = iHeardThat.client.GetAvailableSoundChannel()

		// DebugMsg("I think I sent a [thisSound.volume] volume sound to [iHeardThat.name]")

		iHeardThat << thisSound

/*
	LineOfFire(atom/source,atom/destination, var/range, var/passwall )

	This basically returns a list of turfs that are within the line of fire.
	It expands pon the AbyssDragon getline() code by using the slope to draw the line the length of the range variable.
	If will shorten the line if range is less than the distance between source and destination.

	Passwall determines what happens if a wall is encountered,
	if 0 - stop at wall
	if 1 - go through if it's not an opaque wall (plexweave or some other transparent wall)
	if 2 - go through regardless.

	IMPORTANT: The first step of the path, that of the firer, is removed if source is not at the destination
*/

var/const/LINE_OF_FIRE_DEFAULT_RANGE_TO_CALCULATE = 10

proc/LineOfFire(var/atom/source,var/atom/destination, var/range, var/passwall)

	if (destination == null || source == null)
		DebugMsg("LineOfFire is returning null because destination or source was null.")
		return(null)

	if (passwall == null)
		passwall = 0

	if (range == null)

		range = get_dist(source,destination)

		if (range < 2)
			range = LINE_OF_FIRE_DEFAULT_RANGE_TO_CALCULATE
	else
		range = round(range) // We're going to need a whole number for this one.

	// LineOfFire(source,destination,10,2)

	var/incrementExamined = 1 // increment in lineOfFire we're currently examining.
	var/list/turf/lineOfFire = getline(source,destination)

	if (lineOfFire.len <= 1) // This means the firer is trying to fire into their own hex.  No good.
		// This can actually be normal behavior if the player is goofing off.
		// DebugMsg("LineOfFire() returning null because the firer is attempting to fire into their own hex.")
		return(null)
	// Else, this must be a length of 2 or greater.
	while (lineOfFire.len < (range + 1)) // Plus one because we're hacking off the first member when we're done.
		incrementExamined++ // because it was initialized at 1, this is 2 at the start.
		var/turf/previousTurf = lineOfFire[incrementExamined-1]
		var/turf/currentTurf = lineOfFire[incrementExamined]
		var/directionFlow = get_dir(previousTurf,currentTurf)
		var/turf/newTurf = get_step(lineOfFire[lineOfFire.len],directionFlow)

			// In the event the turf is null, we've reached the edge of the map.
		if (isnull(newTurf))
			break

		lineOfFire.Add(newTurf) // It's a good new turf to add.

			// Now, depending on whether or not we pass walls, we either end it here or keep going.
		switch (passwall)
			if (0) // Pass no wall.
				if (currentTurf.density) // This might be quicker than checking if it's type wall.
					break
			if (1) // Pass transparent (!opacity) walls.
				if (currentTurf.density && currentTurf.opacity)
					break
			//else  Pass all.

	lineOfFire.Cut(1,2) // remove the firer from the firing line.
	return(lineOfFire)

/* ShiftImageInItsDirection()
    Basically sets the pixel_x and pixel_y values of a given image to a value corresponding to its
   dir variable. */
proc/ShiftImageInItsDirection(var/image/thisImage)

	var/const/IMAGE_SHIFT_AMOUNT = 16

	switch (thisImage.dir)
		if (NORTH)
			thisImage.pixel_y = IMAGE_SHIFT_AMOUNT
		if (NORTHEAST)
			thisImage.pixel_x = IMAGE_SHIFT_AMOUNT
			thisImage.pixel_y = IMAGE_SHIFT_AMOUNT
		if (EAST)
			thisImage.pixel_x = IMAGE_SHIFT_AMOUNT
		if (SOUTHEAST)
			thisImage.pixel_x = IMAGE_SHIFT_AMOUNT
			thisImage.pixel_y = -IMAGE_SHIFT_AMOUNT
		if (SOUTH)
			thisImage.pixel_y = -IMAGE_SHIFT_AMOUNT
		if (SOUTHWEST)
			thisImage.pixel_x = -IMAGE_SHIFT_AMOUNT
			thisImage.pixel_y = -IMAGE_SHIFT_AMOUNT
		if (WEST)
			thisImage.pixel_x = -IMAGE_SHIFT_AMOUNT
		if (NORTHWEST)
			thisImage.pixel_x = -IMAGE_SHIFT_AMOUNT
			thisImage.pixel_y = IMAGE_SHIFT_AMOUNT

/* ShiftImageInItsOppositeDirection()
	It so happens that ShiftImageInItsDirection() backfires if you're queing a backwards step.  This is the solution. */
proc/ShiftImageInItsOppositeDirection(var/image/thisImage)

	ShiftImageInItsDirection(thisImage)
	thisImage.pixel_x = thisImage.pixel_x * -1
	thisImage.pixel_y = thisImage.pixel_y * -1



/* Converting things to magntitude was ultimately a limiting mistake.  Maybe I want a 100 energy bolt, for example?
/* MagnitudeToEnergy()
	Converts an amount of magnitude to the energy needed to generate it. */
proc/MagnitudeToEnergy(var/magnitudeToConvert)
	if (magnitudeToConvert == null)
		DebugMsg("null passed to MagnitudeToEnergy.")
		return(0)

	switch (magnitudeToConvert)
		if (0)
			return(0)
		if (1)
			return(1)
		if (2)
			return(2)
		if (3)
			return(4)
		if (4)
			return(8)
		if (5)
			return(16)
		else
			DebugMsg("MagnitudeToEnergy(): Unknown magntiude amount [magnitudeToConvert]")
			return(0)

/* EnergyToMagnitude()
	Converts an amount of energy to the magnitude it would generate. */
proc/EnergyToMagnitude(var/energyToConvert)
	if (energyToConvert == null)
		DebugMsg("null passed to MagnitudeToEnergy.")
		return(0)

	switch (energyToConvert)
		if (0)
			return(0)
		if (1)
			return(1)
		if (2 to 3)
			return(2)
		if (4 to 7)
			return(3)
		if (8 to 15)
			return(4)
		if (16 to 999)
			return(5)
		else
			DebugMsg("EnergyToMagnitude(): Unknown energy amount [energyToConvert]")
			return(0)
*/
