/*
	Project Xenoverse
	By Geldon Yetichsky

file:
Turfs.dm

notes:
	The only turf defined here is the generic world turf /turf/soil, which is handled differently depending on whether it was
	created on the surface or in a base.  Aside from that, this file is pretty much used to handle general turf procs which
	apply to all kinds of turfs.

	Because space, planets, and personal maps are handled so differently, they have completely different procs defined on them.
	The procs here on the /turf/ are mostly used to differentiate which turf this one is.
*/

var/const/SOIL_STEP_NOISE = 'SoilStep.wav'
var/const/PLEXWEAVE_STEP_NOISE = 'PlexweaveStep.wav'
var/const/ALCHELITE_STEP_NOISE = 'AlcheliteStep.wav'
var/const/SUBNEUTRONIUM_STEP_NOISE = 'SubneutroniumStep.wav'
var/const/TRANSMATRIX_STEP_NOISE = 'TransmatrixStep.wav'

turf

	Enter(var/atom/movable/thisObject)

		for (var/tmp/obj/device/shieldgen/thisShield in contents)
			if (thisShield.IsAccessPermittedFor(thisObject))
				continue
			else
				return(0)
		// Otherwise, we go by the default BYOND rules
		return(..())

	Entered(var/atom/movable/thisObject)

		if (istype(thisObject,/mob/player))
			for (var/tmp/obj/device/spawnchamber/thisChamber in contents)
				thisChamber.Dematerialize(thisObject)

		..()


	Click()
		if (usr.client != null)
			usr.client.statobj = src
		..()

	DblClick()
		if (istype(usr,/mob/player))
			var/tmp/mob/player/thisPlayer = usr
			if (thisPlayer.client.activeAbility == null)
				return()
			/* Line of sight check
			if (!(src in view(thisPlayer)))
				GameOutput("[thisPlayer.client.activeAbility.name] cannot be done there: it is out your line of sight.")
				return(0)
			*/
			// thisPlayer.NewAction("UseAbility",src,thisPlayer.client.activeAbility)

			thisPlayer.client.activeAbility.Activate(thisPlayer,src)

		..()

	Del()
		/* So, a great deal of difficulty was overcome when I realized I needed this spawn() here.  The thing is, deleting a
		wall will turn it into another wall or a floor, but I don't know which until it's done deleting itself.  I can't call
		DrawFloorToWallOverlays yet, either, because it's not done deleting itself and being replaced by the next turf yet.
		Thus, spawning a process that waits until this is done deleting is exactly what I needed to do.  Then, it runs
		tDrawFloorToWallOverlays which not only draws needed overlays but removes unneeded ones just in case they're there.

		I run this on Del() instead of New() because it preserves a great deal of world initializaiton time that way.
		On turf/floor.New() I draw the initial walls, and this is relatively harmless because there's quite a few turf/floors
		in the game relative to the walls. */

		spawn()
			sleep()
			var/tmp/turf/thisTurf

			thisTurf = get_step(src,NORTH)
			if (thisTurf != null)
				thisTurf.DrawFloorToWallOverlays()
			thisTurf = get_step(src,EAST)
			if (thisTurf != null)
				thisTurf.DrawFloorToWallOverlays()
			thisTurf = get_step(src,WEST)
			if (thisTurf != null)
				thisTurf.DrawFloorToWallOverlays()
			thisTurf = get_step(src,SOUTH)
			if (thisTurf != null)
				thisTurf.DrawFloorToWallOverlays()
		..()

	/* IsOccupied()
		A quick and dirty way to find out if another solid thing should be able to enter here or not without actually
		knowing what that thing is. */
	proc/IsOccupied()
		if (src.density)
			return(1)
		for (var/atom/thisAtom in src)
			if (istype(thisAtom,/area))
				continue
			else
				if (thisAtom.density == 1)
					return(1)
		return(0)

	/* GetObstruction()
		GetObstruction() is similar to IsOccupied() except it actually returns the name of whatever it is that would be an
		obstructing thing. */
	proc/GetObstruction()
		for (var/atom/thisAtom in src)
			if (istype(thisAtom,/area))
				continue
			else
				if (thisAtom.density == 1)
					return(thisAtom.name)
		return("nothing")

	proc/Beautify()
		if (istype(src,/turf/floor))
			var/turf/floor/thisTurf = src
			thisTurf.DrawFloorToWallOverlays()


	/*
		When called, it generates a sound for stepping on this turf.
	*/
	proc/MakeTurfNoise(var/newVolume)
		if (newVolume == null)
			newVolume = 100

		switch (GetDensity())
			if (0) // Soil
				Noise(src,SOIL_STEP_NOISE,newVolume)

			if (1) // Plexweave
				Noise(src,PLEXWEAVE_STEP_NOISE,newVolume)

			if (2) // Alchelite
				Noise(src,ALCHELITE_STEP_NOISE,newVolume)

			if (3) // Subneutronium
				Noise(src,SUBNEUTRONIUM_STEP_NOISE,newVolume)

			else // Transmatrix
				Noise(src,TRANSMATRIX_STEP_NOISE,newVolume)




	/* DrawFloorToWallOverlays()
		Using overlays to draw wall edges is a powerful, but tricky beast.  This is intended to be called when floors are first
		created or when walls are deleted.  The overlays only are intended to exist on the edge of floor tiles, and I must wait
		until the walls are fully deleted before evaluation can take place.  Thus, for the wall deletion call, I must spawn
		a process to wait until the wall is deleted and then call this on each adjacent space.  The checking to see if the
		space this is being called on is a wall or not is done here because the overlays should only be drawn on floors. */
	proc/DrawFloorToWallOverlays()

		if (istype(src,/turf/wall))
			return()


		var/tmp/thisTurf
		var/directionToCheck

		for (var/increment,increment < 5, increment++)
			switch (increment)
				if (1)
					directionToCheck = NORTH
				if (2)
					directionToCheck = WEST
				if (3)
					directionToCheck = EAST
				else
					directionToCheck = SOUTH

				// Clean up any old overlays.

			overlays.Remove(icon('TurfOverlays.dmi',"wall",directionToCheck))
			overlays.Remove(icon('TurfOverlays.dmi',"wallrough",directionToCheck))

				// Check to see if new overlays need to be applied.

			thisTurf = get_step(src,directionToCheck)

			if (istype(thisTurf,/turf/wall/soil))
				overlays.Add(icon('TurfOverlays.dmi',"wallrough",directionToCheck))
			else if (istype(thisTurf,/turf/wall))
				overlays.Add(icon('TurfOverlays.dmi',"wall",directionToCheck))

	proc/GetDensity()
		switch (type)
			if (/turf/floor/soil)
				return(0)
			if (/turf/wall/soil)
				return(0)
			if (/turf/floor/plexweave)
				return(1)
			if (/turf/wall/plexweave)
				return(1)
			if (/turf/floor/alchelite)
				return(2)
			if (/turf/wall/alchelite)
				return(2)
			if (/turf/floor/subneutronium)
				return(3)
			if (/turf/wall/subneutronium)
				return(3)
			else // Probably transmatrix or some material which doesn't want to be manipulated
				return(4)

	/*
		GetVolumeMuffled()
		This returns a number of volume to remove from a sound.volume call.
		Of course, if we wanted to get super realistic, the volume muffled would probably be somewhat exponential,
		but this simply returns a sequential value representing a percentage of volume to remove
	*/
	var/const/TURF_SOUND_MUFFLED_PER_GETDENSITY_POINT = 20
	proc/GetVolumeMuffled()
		if (density == 0) // Floors do not block sound
			return(0)
		return(GetDensity() * TURF_SOUND_MUFFLED_PER_GETDENSITY_POINT)

	/* Turf.Damage(Attack)
		This is for handling actual attacks aimed at destroying the current turf. */
	proc/Damage(var/Attack/thisAttack)
		if (thisAttack == null)
			DebugMsg("turf/Damage: null thisAttack passed.")
			return(0)

		if (!(istype(thisAttack,/Attack)))
			DebugMsg("turf/Damage arg passed must be of Attack type, not [type].")
			return(0)

		var/tmp/density = GetDensity()
		var/tmp/attackDamage = thisAttack.power // attackDamage is what's ultimately put on the damage tracking object.

		switch (density)
			if (0) // soil type, no actual damage happens to soil
				return(1)
			if (1) // Plexweave: absorbs 4 damage or less, takes double damage.
				attackDamage = rand(1,attackDamage) - 1
			if (2) // Alchelite: absorbs 8 damage or less, takes same damage
				attackDamage = rand(1,attackDamage) - 2
			if (3) // Subneutronium: absorbs 12 damage or less, takes half damage
				attackDamage = rand(1,attackDamage) - 3
			if (4) // This terrain cannot be damaged.
				for (var/tmp/obj/turfdamage/findDamage in contents)
					del findDamage
				return()
			else
				DebugMsg("turf.Damage() unhandled density type: [density]")
				return(0)

		if (attackDamage != null)
			attackDamage = max(attackDamage,0) // Make sure attackDamage is at least 0

			if (attackDamage == 0) // Damage was deflected.
				return(1)
			else
				F_damage(src, attackDamage, "#FF0000")

		var/tmp/obj/turfdamage/thisDamage = null

		for (var/tmp/obj/turfdamage/findDamage in contents)
			if (thisDamage == null)
				thisDamage = findDamage
			else // Getting rid of any redundant damage objects.
				del findDamage

		// This creates a new /obj/turfdamage if needed, otherwise uses the existing one.
		if (thisDamage == null)
			thisDamage = new/obj/turfdamage(src,attackDamage)
		else
			thisDamage.AddDamage(attackDamage)

	proc/DestroyTurf()
		switch (type)
			if (/turf/floor/soil)
				return(0)
			if (/turf/wall/soil)
				return(0)
			if (/turf/floor/plexweave)
				new/turf/floor/soil(src)
			if (/turf/wall/plexweave)
				new/turf/floor/plexweave(src)
			if (/turf/floor/alchelite)
				new/turf/floor/soil(src)
			if (/turf/wall/alchelite)
				new/turf/floor/alchelite(src)
			if (/turf/floor/subneutronium)
				new/turf/floor/soil(src)
			if (/turf/wall/subneutronium)
				new/turf/floor/subneutronium(src)
			else
				DebugMsg("turf/DestroyTurf(): Unknown turf type: [type]")

/*
		default -- Should not actually be substantiated in the game except as a default turf.
		As soon as the New() is run, it is obliterated - replaced with a more appropraite surf.


	default
		name = "Undefined"
		desc = "If you can see this it's probably a bug."

		/* Since the player should never see this turf, this is mostly here just as a map building reference */

		icon = 'TurfsFloor.dmi'
		icon_state = "soil"

		New(var/location)
			// This is where the magic happens of transforming undefined turfs into the appropriate default turf.
			new/turf/wall/soil(src)
			..()
*/