/*
  /obj/device/spawnchamber

  When players are not on the map, they are in the respawnArea (/area/respawn) room, where (among other things) they are
    presented with a list of accessible spawn chambers.

  Without a working spawn chamber (biomaterializer, or whatever) the players simply cannot enter the game.  The proc that
    actually moves the players from the respawn room to the spawn chamber is here, and it will not function if the spawn
    chamber is disabled.

  In order for a spawn chamber to be accessible, the player needs to be aligned to it.
   -- At the most basic level, the spawn chamber and player must be aligned (via GetAlignment())
   -- However, I may require players specifically unlock access to individual spawn chambers.
   --- An easy way to do this is to restrict people who spawn here to those in the group variable.


*/

obj/device/spawnchamber

	var/const/SPAWN_CHAMBER_ENERGY_REQUIRED_TO_RESPAWN = 20
	var/const/MAXIMUM_SPAWN_CHAMBER_ENERGY = 100

	density = 0

	icon = 'imgDevices.dmi'
	icon_state = "spawnchamber"

	name = "Biomaterializer"
	desc = "These biomaterializers are capable of generating the New You (C) bodies in adjacent spaces.  Stepping into them is a handy way to to have your New You (C) recycled for quick travel elsewhere."

	New(var/location,var/initialEnergy)

		maxEnergy = MAXIMUM_SPAWN_CHAMBER_ENERGY

		..(location, initialEnergy)

	DblClick(var/location,var/control,var/params)
		if (usr in respawnArea)
			RespawnPlayer(usr)
		else
			..(location,control,params)

	// Really don't need a custom Heartbeat() or Activate() since the spawnchamber doesn't do much with them but update its appearance

	UpdateAppearance()
		if (icon_state == "spawnchamber")
			if (!IsActivated())
				icon_state = "spawnchamber_deactivated"
		else if (icon_state == "spawnchamber_deactivated")
			if (IsActivated())
				icon_state = "spawnchamber_activated"
		..()

	/* CanRespawnPlayer()
		This proc refers to the check done to see if a player is allowed to respawn here.  Returns 0 if not.
		It does not inform the player why they can't respawn there, it's simply used for populating the respawn list. */
	proc/CanRespawnPlayer(var/mob/player/thisPlayer)

		if (!IsActivated())
			// DebugMsg("CanRespawnPlayer [x] [y]: Can't spawn here, not activated.")
			return(0)

		if (GetEnergy() < SPAWN_CHAMBER_ENERGY_REQUIRED_TO_RESPAWN)
			// DebugMsg("CanRespawnPlayer [x] [y]: Can't spawn here, not enough energy to respawn [energy]/[maxEnergy].")
			return(0)

		if (thisPlayer.GetAlignment() != src.GetAlignment())
			// DebugMsg("CanRespawnPlayer [x] [y]: Can't spawn here: wrong faction. [thisPlayer.GetAlignment()] vrs [src.GetAlignment()]")
			return(0)
			  // Maybe I should take out this check and have it simply message the player if the respawner is obstructed.
			  // On the other hand, knowing the respawners are obstructed would be handy to know we need to spawn more.
		if (FindRespawnSpace() != null)
			return(1)
		else
			// DebugMsg("CanRespawnPlayer [x] [y]: Can't spawn here: no free spaces.")
			return(0)


	/* RespawnPlayer()
		This proc is actually called when the player has the spawn chamber selected and clicks the respawn button on the
		respawn screen and has this respawn chamber selected. */
	proc/RespawnPlayer(var/mob/player/thisPlayer)

		/* This did not work.
		if (thisPlayer.loc != /area/respawn)
		*/
		if (thisPlayer.IsAlive())
			// This is really an error state.  The player shouldn't even be able to *see* the respawn button at this point.
			thisPlayer.GameOutput("You cannot be revived: you are still alive.")
			return()

		//  First, check that the player is allowed to spawn here.

		if (thisPlayer.GetAlignment() != src.GetAlignment())
			thisPlayer.GameOutput("Access denied.")
			return()

		if (GetEnergy() < SPAWN_CHAMBER_ENERGY_REQUIRED_TO_RESPAWN)
			thisPlayer.GameOutput("There's not enough energy to rematerialize there.")
			return()

		// Next, perform all the work of spawning the player here.

		var/turf/turfToRespawnAt = FindRespawnSpace()

		if (turfToRespawnAt == null)
			thisPlayer.GameOutput("There's no room to rematerialize there.")
			return()

		thisPlayer.density = 1 // Setting this here in order to make sure the Move() check is performed correctly.

		if (thisPlayer.Move(turfToRespawnAt))
			thisPlayer.dir = get_dir(src,turfToRespawnAt)
			// This would be a good spot for FX related to rematerializing players.
			new/obj/effects/crackling(turfToRespawnAt)
			sleep(10)
			thisPlayer.ActionRespawned()
		else
			thisPlayer.GameOutput("Technical difficulties have prevented you from rematerializing there.")
			thisPlayer.density = 0 // If you're still in /area/respawn, there you should remain.
			return()

		// Lastly, don't forget to cause whatever ramfications would impact this spawnchamber.

		AdjustEnergy(-SPAWN_CHAMBER_ENERGY_REQUIRED_TO_RESPAWN)

	/* FindRespawnSpace()
		Returns null if there is no respawn space here (perhaps something is blocking it) otherwise returns a random one
		of the viable places found to respawn. */
	proc/FindRespawnSpace()

		var/list/spawnTurfs = getring(src,1)

		while (spawnTurfs.len > 0)
			var/turf/thisTurf = pick(spawnTurfs)
			if (thisTurf == null || thisTurf.IsOccupied())
				spawnTurfs.Remove(thisTurf)
				continue
			return(thisTurf)
		return(null)

	/* Dematerialize()
		Called when a player enters the Biomaterializer's location.
		If they are on the same faction, they are sent back to the respawn screen.
		Otherwise... well, killing them seems sort of unsporting, lets just deny their access. */

	proc/Dematerialize(var/mob/player/thisPlayer)

		if (!IsActivated())
			thisPlayer.GameOutput ("You step onto the [name], but it appears to be deactivated.")
			return()

		if (thisPlayer.GetAlignment() == GetAlignment())
			thisPlayer.Move(respawnArea)
			new/obj/effects/crackling(src.loc)
		else
			thisPlayer.GameOutput ("You step onto the [name], but nothing happens.")


/*

					if (prestige == 0)
						GameOutput("You find yourself in a white room, dripping wet, apparently having fallen out of some kind of floating vat behind you.  You're wearing some kind of suit but have nothing else.  Your memory is fuzzy, but you seem to recall you're in some kind of danger.",src)
					else
						GameOutput("You emerge from the spawn chamber, soaking wet, battered, but alive.",src)
*/